/* FIN2BMP.C

   Copyright (c)2002 Andrew Durdin. Email: andy@durdin.net

   You have permission to use this code in your own projects, on the conditions
   that leave this notice intact, and that you give proper credit in your final
   program. It'd be nice if you let me know what you do with it, also.
*/

#include <stdio.h>
#include <string.h>

#include "bmp16.h"


int main( int argc, char *argv[] ) {
	int result = 0;

	if( argc < 4 ) {
	  	printf( "Use: FIN2BMP -b finale.ck? output.bmp\n"
				"  or FIN2BMP -f input.bmp finale.ck?\n\n"
 				"You can also use this util for changing\n"
 				"preview1.ck1 and preview2.ck1\n\n" );
		exit( 1 );
	}

	if( stricmp( argv[1], "-b" ) == 0 ) {
		result = fin_to_bmp( argv[2], argv[3] );
	} else if( stricmp( argv[1], "-f" ) == 0 ) {
		result = bmp_to_fin( argv[2], argv[3] );
	} else {
		printf( "Invalid parameter %s!\n", argv[1] );
		result = 1;
	}

	return result;
}

int fin_to_bmp( char *finfile, char *bmpfile ) {
	FILE *fin;
	BITMAP16 *bmp;
	int result = 0;
	int x, y, plane, num;
	int b, bitmask, i;
	long count, maxcount;
	unsigned char bytes[130];
	unsigned char bytecount;

	fin = fopen( finfile, "rb" );
	if( !fin ) {
		printf( "Cannot open input file for reading!\n" );
	  	return 1;
	}

	bmp = bmp_create( 320, 200 );
	if( !bmp ) {
		printf( "Can't create bitmap!\n" );
	  	fclose( fin );
		return 1;
	}

	/* Clear the bitmap */
	bmp_rect( bmp, 0, 0, 319, 199, 0 );

	/* Read the plane data size from the file */
	fread( &maxcount, sizeof( long ), 1, fin );

	printf( "Decompressing...\n" );

	count = x = y = 0;
	plane = 1;
	while( (plane <= 8) && (b = fgetc( fin )) != EOF ) {
		if( b & 0x80 ) {
	    	//N + 1 bytes of data follows
	        bytecount = (b & 0x7F) + 1;
			fread( bytes, bytecount, 1, fin );
		} else {
            //Repeat N + 3 of following byte
           	bytecount = b + 3;
			b = fgetc( fin );
			memset( bytes, b, bytecount );
		}

		/* Draw buffered bits */
		for( i = 0; i < bytecount; i++ ) {
			for( bitmask = 0x80; bitmask > 0; bitmask >>= 1 ) {
				if( bytes[i] & bitmask )
				  	bmp_putpixel( bmp, x, y, bmp_getpixel( bmp, x, y ) | plane );
		        x++;
				if( x >= 320 ) {
				  	x = 0;
					y++;
				}
				count++;
				if( count >= maxcount*2 ) {
				  	count = 0;
					x = 0;
				  	y = 0;
					plane <<= 1;
				}
            }
	    }
	}

	/* The BMP now contains the picture--save it and return */
	if( !bmp_save( bmp, bmpfile ) ) {
		result = 1;
		printf( "Can't save bitmap file!\n" );
	} else
		printf( "Done!\n" );

	bmp_free( bmp );
	fclose( fin );

	return result;
}


int bmp_to_fin( char *bmpfile, char *finfile ) {
	unsigned char *datstart, *runstart, *probe;
	long count;
	int plane, len;

	FILE *fout;
	BITMAP16 *bmp;

	/* Load the bmp, and make sure it's the right size */
	bmp = bmp_load( bmpfile );
	if( !bmp ) {
		printf( "Can't find input bitmap, or not 16 colours!\n" );
		return 1;
	}

	if( bmp->w != 320 || bmp->h != 200 ) {
		printf( "Input bitmap should be 320x200 16 colours!\n" );
		bmp_free( bmp );
		return 1;
	}

	/* Load the input data file */
	fout = fopen( finfile, "wb" );
	if( !fout ) {
		printf( "Can't open output file for writing!\n" );
		bmp_free( bmp );
		return 1;
	}

	/* Unpack the bmp */
	bmp_unpack( bmp );


	/* Write the plane data size from the file */
	count = 0x8000L;
	fwrite( &count, sizeof( long ), 1, fout );

	printf( "Compressing...\n" );

	/* Compress the bits in each plane into the file */
	for( plane = 0; plane < 4; plane++ ) {

		count = 8000;
		datstart = bmp->bits + plane*bmp->linew*bmp->h;
		runstart = probe = datstart;

		while( count ) {
			/* Probe for the start of a run */
			while( count && (probe - runstart) < 3 ) {
				probe++; count--;
				if( *probe != *runstart )
					runstart = probe;
			}
			/* Probe for the end of the run */
			while( count && *probe == *runstart ) {
				probe++; count--;
			}

			/* Now we should have datstart the start of the data, */
			/* runstart the start of the run, and probe 1 past the */
			/* end of the run */

			/* Save the data */
			while( (runstart - datstart) > 0 ) {
				len = (runstart - datstart);
				if( len > 128 ) len = 128;

				/* Write a block of data */
				fputc( (len - 1) | 0x80, fout );
				fwrite( datstart, len, 1, fout );
				datstart += len;
			}

			/* Save the run */
			while( (probe - runstart) > 0 ) {
				len = (probe - runstart);
				if( len > 130 ) len = 130;

				/* Write a run code */
				fputc( (len - 3), fout );
				fputc( *runstart, fout );
				runstart += len;
			}

			/* And continue from where the probe is */
			datstart = runstart = probe;
		}

		/* output 192 extra padding bytes at the end of each plane */
		fputc( 0x7D, fout );
		fputc( 0, fout );
		fputc( 0x3D, fout );
		fputc( 0, fout );

	}

	fclose( fout );

	bmp_free( bmp );

	printf( "Done.\n" );

	return 0;
}

